//////// ================================
//////// SETTINGS: Settings
//////// ================================
//////// Most of the interesting code and settings are in Buffer A.
//////// This file consists only of postprocessing.

// The dithering looks *fantastic*. Honestly, at 1440p it makes even
// 3 bits per color channel look *very* convincing. As far as I'm aware
// there are no downsides to leaving it enabled. It would probably help
// reduce color banding and improve color clarity in general.
#define ENABLE_DITHER

// The number of available colors *per channel*.
const uint DITHER_COLORS = uint(1<<8);

// The size of the Bayer matrix is 2^DITHER_BASE, so e.g. 4 is a 16x16 matrix.
// This can't be larger than 16 because dithering is implemented using a 16-bit bit hack.
const uint DITHER_BASE = uint(16);

// Artifically restrict the colors to those specified in DITHER_COLORS.
// If you set DITHER_COLORS to 6 or so and enable DITHER_NEAREST then screenshot
// and convert the image to a GIF* using an appropriate conversion tool,
// it will look exactly the same as it looks on your screen.
// This was the motivating reason for adding dithering, and it looks amazing.
//
// * I pronounce it "yif". Fite me.
//#define DITHER_NEAREST

//////// ================================
//////// IMPL: Implementation
//////// ================================

/// Convert a color from linear RGB to the sRGB color space.
vec3 linear2srgb(vec3 color);

vec4 dither(uvec2 coord, vec4 color);
vec4 nearest_color(vec4 color);

void mainImage(out vec4 fragColor, in vec2 fragCoord) {
    vec2 uv = fragCoord/iResolution.xy;
    uvec2 coord = uvec2(fragCoord); // for dithering
    vec4 color = texture(iChannel0, uv);
    color.rgb = linear2srgb(color.rgb);
    #ifdef ENABLE_DITHER
    color = dither(coord, color);
    #endif
    #ifdef DITHER_NEAREST
    color = nearest_color(color);
    #endif
    fragColor = color;
}

//////// --------------------------------
//////// DITHER: Ordered dithering
//////// --------------------------------
//////// https://en.wikipedia.org/wiki/Ordered_dithering

const uint DITHER_SIZE = uint(1)<<DITHER_BASE;
const uint BIT_WIDTH = DITHER_BASE;

vec4 nearest_color(vec4 color) {
    return floor(color * float(DITHER_COLORS)) / float(DITHER_COLORS);
}

uint bit_reverse(uint x) {
    uint hi = uint(1 << BIT_WIDTH-uint(1));
    uint lo = uint(1);
    for (uint i = uint(0); i < BIT_WIDTH/uint(2); i++) {
        uint bit_hi = x & hi;
        uint bit_lo = x & lo;
        x &= ~hi & ~lo;
        if (bit_hi > uint(0)) x |= lo;
        if (bit_lo > uint(0)) x |= hi;
        hi >>= 1;
        lo >>= 1;
    }
    return x;
}

uint bit_interleave(uint x, uint y) {
    uint mask = uint(1) << BIT_WIDTH-uint(1);
    uint acc = uint(0);
    for (uint i = uint(0); i < BIT_WIDTH; i++) {
        acc |= (x & mask) << uint(2)*i + uint(1);
        acc |= (y & mask) << uint(2)*i;
        mask >>= 1;
    }
    return acc;
}

float bayer(uvec2 coord) {
    // magic bitwise formula from Wikipedia
    uint magic = bit_reverse(bit_interleave(coord.x ^ coord.y, coord.x));
    return float(magic+uint(1)) / (float(DITHER_SIZE)*float(DITHER_SIZE)) - 0.5;
}

vec4 dither(uvec2 coord, vec4 color) {
    if (DITHER_SIZE < uint(2)) return color;
    coord %= DITHER_SIZE;
    vec4 bias = vec4(
        bayer(coord),
        bayer(uvec2(uint(DITHER_SIZE) - coord.x - uint(1), coord.y)),
        bayer(uvec2(coord.x, uint(DITHER_SIZE) - coord.y - uint(1))),
        bayer(uvec2(uint(DITHER_SIZE) - coord.x - uint(1), uint(DITHER_SIZE) - coord.y - uint(1)))
    );
    return color + (bias / float(DITHER_COLORS));
}


//////// ================================
//////// VENDOR: Vendored code
//////// ================================

////
//// AUTHOR: unknown
////

vec3 linear2srgb(vec3 linear_rgb) {
    // I believe the first version is technically more accurate,
    // but the difference is usually negligable in practice.
    #if 1
    // copied from somewhere on the internet
    bvec3 cutoff = lessThan(linear_rgb, vec3(0.0031308));
    vec3 higher = vec3(1.055)*pow(linear_rgb, vec3(1.0/2.4)) - vec3(0.055);
    vec3 lower = linear_rgb * vec3(12.92);

    return mix(higher, lower, cutoff);
    // end copied from somewhere on the internet
    #else
    return pow(linear_rgb, vec3(1./2.2));
    #endif
}
